import logging
from binascii import crc32
from enum import IntEnum
from paramiko import Channel
from queue import Empty, Queue
from selectors import SelectSelector, EVENT_WRITE
from socket import socket, error as SocketError
from struct import pack, unpack
from threading import Event, Thread
from time import sleep
from typing import Any, NamedTuple, Set, Tuple, Union

from .config import TIMEOUT, LEN_HEAD

Connection = Union[socket, Channel]


class Flag(IntEnum):
    PUSH = 1         # 推送申请
    PULL = 2         # 拉取申请
    SID = 3          # 建立会话
    ATTACH = 4       # 后续连接
    MONOFILE = 5     # 传输模式
    DIR_INFO = 6     # 目录信息
    FILE_INFO = 7    # 文件信息
    FILE_COUNT = 8   # 文件数量
    FILE_READY = 9   # 文件就绪
    FILE_CHUNK = 10  # 数据传输
    DONE = 11        # 完成
    EXCEPTION = 12   # 异常退出

    @classmethod
    def contains(cls, member: object) -> bool:
        return member in cls.__members__.values()


class Packet(NamedTuple):
    flag: Flag
    body: bytes

    def __str__(self) -> str:
        return (f'Packet: {self.flag.name} '
                f'len={self.length} '
                f'chk={self.chksum:08x}')

    @property
    def length(self) -> int:
        return len(self.body)

    @property
    def chksum(self) -> int:
        return crc32(self.body)

    @staticmethod
    def load(flag: Flag, *args) -> 'Packet':
        '''将包体封包'''
        if flag == Flag.PULL or flag == Flag.PUSH:
            if isinstance(args[0], bytes):
                body = args[0]
            else:
                body = str(args[0]).encode('utf8')
        elif flag == Flag.SID or flag == Flag.ATTACH:
            body = pack('>16s', *args)
        elif flag == Flag.MONOFILE:
            body = pack('>?', *args)
        elif flag == Flag.DIR_INFO:
            length = len(args[-1])
            body = pack(f'>IH{length}s', *args)
        elif flag == Flag.FILE_INFO:
            length = len(args[-1])
            body = pack(f'>IHQd16s{length}s', *args)
        elif flag == Flag.FILE_COUNT:
            body = pack('>I', *args)
        elif flag == Flag.FILE_READY:
            body = pack('>I', *args)
        elif flag == Flag.FILE_CHUNK:
            length = len(args[-1])
            body = pack(f'>2I{length}s', *args)
        elif flag == Flag.DONE:
            body = pack('>?', True)
        elif flag == Flag.EXCEPTION:
            body = str(args[0]).encode('utf8')
        else:
            raise ValueError(f'{flag} is not a valid Flag')
        return Packet(flag, body)

    def pack(self) -> bytes:
        '''封包'''
        fmt = f'>BIH{self.length}s'
        return pack(fmt, self.flag, self.chksum, self.length, self.body)

    @staticmethod
    def unpack_head(head: bytes) -> Tuple[Flag, int, int]:
        '''解析 head'''
        flag, chksum, length = unpack('>BIH', head)
        if not Flag.contains(flag):
            raise PacketError
        else:
            return Flag(flag), chksum, length

    def unpack_body(self) -> Tuple[Any, ...]:
        '''将 body 解包'''
        if self.flag == Flag.PULL or self.flag == Flag.PUSH:
            return (self.body.decode('utf-8'),)  # connection info

        elif self.flag == Flag.SID or self.flag == Flag.ATTACH:
            return unpack('>16s', self.body)  # Worker ID

        elif self.flag == Flag.MONOFILE:
            return unpack('>?', self.body)  # is monofile

        elif self.flag == Flag.DIR_INFO:
            # file_id | perm | path
            #   4B    |  2B  |  ...
            fmt = f'>IH{self.length - 6}s'
            return unpack(fmt, self.body)

        elif self.flag == Flag.FILE_INFO:
            # file_id | perm | size | mtime | chksum | path
            #   4B    |  2B  |  8B  |  8B   |  16B   |  ...
            fmt = f'>IHQd16s{self.length - 38}s'
            return unpack(fmt, self.body)

        elif self.flag == Flag.FILE_COUNT:
            return unpack('>I', self.body)  # file count

        elif self.flag == Flag.FILE_READY:
            return unpack('>I', self.body)  # file id

        elif self.flag == Flag.FILE_CHUNK:
            # file_id |  seq  | chunk
            #    4B   |  4B   |  ...
            fmt = f'>2I{self.length - 8}s'
            return unpack(fmt, self.body)

        elif self.flag == Flag.DONE:
            return unpack('>?', self.body)

        elif self.flag == Flag.EXCEPTION:
            return (self.body.decode('utf-8'),)

        else:
            raise ValueError(f'{self.flag} is not a valid Flag')

    def is_valid(self, chksum: int):
        '''是否是有效的包体'''
        return self.chksum == chksum


class PacketError(Exception):
    pass


def send_pkt(conn: Connection, packet: Packet):
    '''发送数据报文'''
    datagram = packet.pack()
    conn.sendall(datagram)


def recv_all(conn: Connection, length: int) -> bytes:
    '''接受完整数据'''
    datagram = bytearray()
    while length > 0:
        _data = conn.recv(length)
        n_recv = len(_data)
        if n_recv > 0:
            length -= n_recv
            datagram += _data
        else:
            raise ConnectionResetError

    return bytes(datagram)


def recv_pkt(conn: Connection) -> Packet:
    '''接收数据报文'''
    # 接收并解析 head 部分
    head = recv_all(conn, LEN_HEAD)
    flag, chksum, len_body = Packet.unpack_head(head)

    # 接收 body 部分
    body = recv_all(conn, len_body)
    if crc32(body) == chksum:
        return Packet(flag, body)
    else:
        raise PacketError


class Counter:
    def __init__(self):
        self.n_sent = 0

    def acc(self, length):
        self.n_sent += length


class ConnectionPool(Thread):
    _max_size = 128

    def __init__(self, size=16):
        super().__init__(daemon=True)
        self.size = min(size, self._max_size)
        self.send_q = Queue(self.size)
        self.recv_q = Queue()
        self.done = Event()
        self.sender = SelectSelector()
        self.connections: Set[Connection] = set()

    def send(self, packet: Packet):
        self.send_q.put(packet)

    def recv(self, timeout=TIMEOUT) -> Packet:
        return self.recv_q.get(timeout)

    def add(self, conn: Connection):
        '''添加一个连接'''
        # 检查数量是否达到上限
        if len(self.connections) >= self._max_size:
            return False
        # 检查是否已添加过
        if conn in self.connections:
            return True

        self.connections.add(conn)
        self.sender.register(conn, EVENT_WRITE, data=Counter())

        t_recv = Thread(target=self.listen_to_recv, args=(conn,), daemon=True)
        t_recv.start()
        return True

    def pop(self, conn: Connection):
        try:
            self.sender.unregister(conn)
        except (KeyError, ValueError):
            pass

        try:
            self.connections.remove(conn)
        except KeyError:
            pass
        finally:
            conn.close()

    def listen_to_send(self):
        while not self.done.is_set():
            # find the conn that sent the least data
            keys = [key for key, _ in self.sender.select(timeout=1)]
            if not keys:
                continue
            else:
                key = min(keys, key=lambda k: k.data.n_sent)

            # get data
            try:
                packet: Packet = self.send_q.get(timeout=TIMEOUT)
                conn: Connection = key.fileobj
            except Empty:
                break

            # send
            try:
                send_pkt(conn, packet)
                key.data.acc(packet.length)
            except SocketError as e:
                self.pop(conn)
                logging.warning(f'[Send] Conn-{id(conn):x}: {e}.')

    def listen_to_recv(self, conn: Connection):
        conn_name = f'{id(conn):x}'
        while not self.done.is_set():
            try:
                packet = recv_pkt(conn)
                self.recv_q.put(packet)
                logging.debug(f'[Recv] conn-{conn_name}: {packet}')
            except ConnectionResetError:
                self.pop(conn)
                return
            except SocketError as e:
                self.pop(conn)
                logging.warning(f'[Recv] Conn-{conn_name}: {e}.')
            except PacketError:
                self.pop(conn)
                logging.error(f'conn-{conn_name} received an error packet.')
                return

    def stop(self):
        while not self.send_q.empty():
            sleep(0.2)
        self.done.set()
        self.sender.close()
        for conn in self.connections.copy():
            conn.close()

    def run(self):
        if not self.connections:
            raise ValueError('No connection')

        self.done.clear()
        self.listen_to_send()
        self.stop()



#!/usr/bin/env python

import os
import sys
import logging
import signal
from argparse import ArgumentParser, RawDescriptionHelpFormatter
from functools import partial, wraps
from getpass import getpass, getuser
from json import dumps
from os.path import abspath
from pwd import getpwuid
from socket import create_connection
from textwrap import dedent
from threading import Thread
from time import sleep
from typing import Any, Dict, List, Tuple

from paramiko import Channel, Transport, SSHConfig
from paramiko import RSAKey, DSSKey, ECDSAKey, Ed25519Key
from paramiko import SSHException, ChannelException, PasswordRequiredException
from rich.live import Live
from rich.progress import Progress, SpinnerColumn, TextColumn
from rich.table import Table

from .config import SERVER_ADDR, SSH_MUX, TIMEOUT
from .network import Flag, Packet, send_pkt, recv_pkt
from .transfer import Sender, Receiver, trans_progress


conn_progress = Progress(
    TextColumn("[bold blue]Connecting "),
    TextColumn("[progress.percentage]{task.completed}/{task.total} "),
    SpinnerColumn(finished_text='✓')
)


def retry(num, *, wait, exceptions=(Exception,)):
    def deco(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            for i in range(num):
                try:
                    return func(*args, **kwargs)
                except exceptions as err:
                    if i + 1 == num:
                        raise RuntimeError from err
                    sleep(wait)
                    continue
        return wrapper
    return deco


class Client:
    default_port = 22
    default_dir = os.path.expanduser('~/.ssh')
    default_config = os.path.join(default_dir, 'config')

    def __init__(self, cli_parser: ArgumentParser):
        args = cli_parser.parse_args()

        # init logger
        self.log_level = {
            0: logging.ERROR,
            1: logging.WARNING,
            2: logging.INFO,
            3: logging.DEBUG
        }.get(args.verbose, logging.ERROR)
        self.set_log()

        try:
            # init instance attributes from CLI args:
            #   - self.action:   Flag.PUSH | Flag.PULL
            #   - self.host:     ssh server's ip or domain
            #   - self.username: username used to login to the ssh server
            #   - self.srcs:     source dirs or files
            #   - self.dst:      dest dir or file
            self.parse_cli_args(args)
        except Exception as e:
            msg = f'[b]fcp[/b]: {e}'
            logging.fatal(msg)
            print('-' * (len(msg) - 7))
            cli_parser.print_help()
            sys.exit(1)

        self.config = self.load_ssh_config(self.host, args.ssh_config)
        self.port = args.port
        self.pkey_path = args.private_key
        self.include = args.include
        self.exclude = [p for p in args.exclude.split(',') if p]
        self.n_tunnel = args.num
        self.n_channel = self.n_tunnel * SSH_MUX
        self.conn_tid = conn_progress.add_task('Connecting',
                                               total=self.n_channel)

        # the ssh tunnels
        self.tunnels: Dict[Transport, List[Channel]] = {}

    @staticmethod
    def parse_remote_addr(remote):
        '''解析远程主机登录信息'''
        netloc, path = remote.split(':')
        user, host = netloc.split('@') if '@' in netloc else ('', netloc)
        return user, host, path

    @classmethod
    def parse_remote_sources(cls, sources):
        users, hosts, srcs = set(), set(), set()
        for src in sources:
            user, host, path = cls.parse_remote_addr(src)
            users.add(user)
            hosts.add(host)
            srcs.add(path)
        if len(users) == 1 and len(hosts) == 1:
            return users.pop(), hosts.pop(), sorted(srcs)
        else:
            raise ValueError('All source args must come from '
                             'the same machine with same user.')

    def parse_cli_args(self, args):
        '''解析命令行参数'''
        if ':' in args.srcs[0]:
            self.username, self.host, self.srcs = self.parse_remote_sources(args.srcs)
            self.action = Flag.PULL
            self.dst = abspath(args.dst)
        elif ':' in args.dst:
            self.username, self.host, self.dst = self.parse_remote_addr(args.dst)
            self.action = Flag.PUSH
            self.srcs = [abspath(path) for path in args.srcs]
        else:
            raise ValueError('The server address is not specified.')

    def set_log(self):
        '''处理日志'''
        global print
        print = partial(conn_progress.print, style='cyan')

        logging.root.setLevel(self.log_level)
        if self.log_level <= logging.ERROR:
            logging.fatal = partial(conn_progress.print, style='bold red')
            logging.error = partial(conn_progress.print, style='red')
        if self.log_level <= logging.WARNING:
            logging.warning = partial(conn_progress.print, style='yellow')
        if self.log_level <= logging.INFO:
            logging.info = partial(conn_progress.print, style='blue')
        if self.log_level <= logging.DEBUG:
            logging.debug = partial(conn_progress.print, style='white')

        logging.getLogger("paramiko").setLevel(logging.FATAL)

    @classmethod
    def load_ssh_config(cls, hostname: str, user_config_file=None) -> dict:
        '''加载默认配置'''
        _path = user_config_file or cls.default_config
        if os.path.isfile(_path):
            cfg = SSHConfig.from_path(_path)
            return cfg.lookup(hostname)
        else:
            return {}

    @staticmethod
    def load_pkey(key_path):
        '''加载 Key'''
        _path = os.path.abspath(os.path.expanduser(key_path))
        if not os.path.isfile(_path):
            return

        # guess key type
        filename = os.path.basename(_path)
        key_type = filename.split('_')[1] if filename.startswith('id_') else ''
        key_types = ['rsa', 'ed25519', 'dsa', 'ecdsa']
        key_classes = {
            'rsa': RSAKey,
            'ed25519': Ed25519Key,
            'dsa': DSSKey,
            'ecdsa': ECDSAKey
        }

        types_to_try = [key_type] if key_type in key_types else key_types
        for _ in range(3):
            password = None
            for _type in types_to_try:
                key_cls = key_classes[_type]
                try:
                    return key_cls.from_private_key_file(key_path, password)
                except PasswordRequiredException:
                    password = getpass(f"Enter password for key '{key_path}': ")
                except SSHException:
                    continue

    def search_pkeys(self):
        '''查找可用的私钥'''
        if self.pkey_path:
            pkey_paths = [self.pkey_path]
        else:
            pkey_paths = self.config.get('identityfile', [])

        if not pkey_paths:
            for keyname in os.listdir(self.default_dir):
                if keyname.startswith('id_') and not keyname.endswith('.pub'):
                    path = os.path.join(self.default_dir, keyname)
                    pkey_paths.append(path)

        logging.debug(f'found pkeys: {pkey_paths}')

        return pkey_paths

    @retry(3, wait=0.3, exceptions=(ConnectionResetError, EOFError))
    def create_transport(self, sock, user, pkey, password):
        if isinstance(sock, tuple):
            sock = create_connection(sock)
        tp = Transport(sock)
        tp.use_compression(True)
        tp.set_keepalive(60)
        try:
            tp.connect(username=user, pkey=pkey, password=password)
        except SSHException as e:
            tp.stop_thread()
            if 'Connection reset by peer' in str(e):
                raise ConnectionResetError from e
            else:
                logging.error(f'[b]fcp[/b]: failed to create transport due to {e}')
                return

        if tp.is_authenticated():
            self.tunnels[tp] = []
            return tp
        else:
            logging.error('[b]fcp[/b]: failed to create transport due to authentication')

    def create_channel(self, transport: Transport) -> Channel:  # type: ignore
        '''create a new channel by transport'''
        try:
            # create channel
            channel = transport.open_channel(kind='direct-tcpip',
                                             dest_addr=SERVER_ADDR,
                                             src_addr=('localhost', 0))
            channel.settimeout(TIMEOUT)

            # add to self.tunnels
            channels = self.tunnels.setdefault(transport, [])
            channels.append(channel)

            conn_progress.update(self.conn_tid, advance=1)
            return channel
        except (SSHException, ChannelException) as e:
            logging.error(f'[b]fcp[/b]: create channel failed due to {e}')
            sys.exit(1)

    def ssh_connect(self) -> Tuple[Transport, Any, Any]:  # type: ignore
        '''连接 SSH 服务器'''
        # 获取 SSH 服务器的连接参数
        self.host = self.config.get('hostname', self.host)
        self.port = self.port or self.config.get('port') or self.default_port
        self.port = int(self.port)
        self.username = self.username or self.config.get('user') or getuser()
        logging.debug(f'login info: {self.username}@{self.host}:{self.port}')

        # search private keys
        pkey_paths = self.search_pkeys()

        # connect to ssh server (just connect, not auth)
        addr = (self.host, self.port)

        # try the pkeys one by one
        for _path in pkey_paths:
            logging.info(f'[b]test pkey[/b]: {_path}')
            pkey = self.load_pkey(_path)
            tp = self.create_transport(addr, self.username, pkey, None)
            if tp:
                self.tunnels[tp] = []
                return tp, pkey, None

        # try to auth with password
        for _ in range(3):
            password = conn_progress.console.input(
                f'password for {self.username}@{self.host}:\n',
                password=True
            )
            logging.debug('[b]test password[/b]: %s' % password)
            tp = self.create_transport(addr, self.username, None, password)
            if tp:
                self.tunnels[tp] = []
                return tp, None, password
            else:
                print('[b]fcp[/b]: wrong password!', style='red')  # type:ignore

        logging.error('[b]fcp[/b]: failed to create SSH tunnel')
        sys.exit(1)

    def handshake(self, channel, conn_info: str):
        '''握手'''
        conn_pkt = Packet.load(self.action, conn_info)
        send_pkt(channel, conn_pkt)
        session_pkt = recv_pkt(channel)
        session_id, = session_pkt.unpack_body()
        logging.info(f'[b]fcp[/b]: Channel-{id(channel):x} connected')

        return session_id

    def create_attached_channels(self, tp, conn_pool, session_id):
        channels = self.tunnels[tp]
        attach_pkt = Packet.load(Flag.ATTACH, session_id)

        def _attache_channel():
            channel = self.create_channel(tp)
            send_pkt(channel, attach_pkt)
            conn_pool.add(channel)
            logging.info(f'[b]fcp[/b]: Channel-{id(channel):x} connected')

        for _ in range(SSH_MUX - len(channels)):
            thr = Thread(target=_attache_channel, daemon=True)
            thr.start()

    def attached_connect(self, conn_pool, session_id, pkey, password):
        '''后续连接'''
        addr = (self.host, self.port)

        @retry(3, wait=0.3, exceptions=ConnectionResetError)
        def _attache():
            tp = self.create_transport(addr, self.username, pkey, password)
            if tp:
                self.create_attached_channels(tp, conn_pool, session_id)

        # create channels from exists transports
        for tp in self.tunnels:
            self.create_attached_channels(tp, conn_pool, session_id)

        # create new transports
        for _ in range(self.n_tunnel - len(self.tunnels)):
            thr = Thread(target=_attache, daemon=True)
            thr.start()
            sleep(0.3)

    def start(self):
        progress_table = Table.grid()
        progress_table.add_row(conn_progress)
        progress_table.add_row(trans_progress)

        with Live(progress_table, refresh_per_second=10):
            try:
                tp, pkey, password = self.ssh_connect()
                first_channel = self.create_channel(tp)
                local_user = getpwuid(os.getuid()).pw_name

                if self.action == Flag.PULL:
                    conn_info = dumps({
                        'user': self.username,
                        'srcs': self.srcs,
                        'include': self.include,
                        'exclude': self.exclude
                    }, ensure_ascii=False, separators=(',', ':'))
                    session_id = self.handshake(first_channel, conn_info)
                    porter = Receiver(session_id, local_user, self.dst,
                                      self.n_channel)
                else:
                    conn_info = dumps({
                        'user': self.username,
                        'dst': self.dst
                    })
                    session_id = self.handshake(first_channel, conn_info)
                    porter = Sender(session_id, local_user, self.srcs,
                                    self.n_channel, self.include, self.exclude)

                porter.conn_pool.add(first_channel)
                porter.start()

                # create attached connections
                t = Thread(target=self.attached_connect,
                           args=(porter.conn_pool, session_id, pkey, password))
                t.start()

                porter.join()
            except Exception as e:
                logging.fatal(f'[b]fcp[/b]: {e}')
                if self.log_level == logging.DEBUG:
                    conn_progress.console.print_exception()
                sys.exit(1)


def handle_sigint(signum, frame):
    '''键盘中断事件的处理'''
    logging.error('[b]fcp[/b]: user canceled.')
    conn_progress.stop()
    trans_progress.stop()
    sys.exit(1)


signal.signal(signal.SIGINT, handle_sigint)


def main():
    parser = ArgumentParser(
        prog='fcp',
        formatter_class=RawDescriptionHelpFormatter,
        description=dedent('''
            PULL : fcp [OPTIONS...] [USER@]HOST:SRC... DST
            PUSH : fcp [OPTIONS...] SRC... [USER@]HOST:DST
        ''')
    )

    parser.add_argument('-p', dest='port', type=int, default=None,
                        help='The port of SSH server (default: 22)')

    parser.add_argument('-i', dest='private_key', type=str, default=None,
                        help='The private key file for SSH')

    parser.add_argument('-F', dest='ssh_config', type=str, default=None,
                        help='The config file for SSH (default: ~/.ssh/config)')

    parser.add_argument('-n', dest='num', type=int, default=8,
                        help='Max number of SSH tunnels (default: %(default)s)')

    parser.add_argument('-v', dest='verbose', action='count', default=0,
                        help='Verbose mode (default: disable)')

    parser.add_argument('--include', type=str, metavar='PATTERN', default='*',
                        help='include files matching PATTERN')

    parser.add_argument('--exclude', type=str, metavar='PATTERN', default='',
                        help='exclude files matching PATTERN, split by `,`')

    parser.add_argument(dest='srcs', nargs='+', help='source path')
    parser.add_argument(dest='dst', help='destination path')

    cli = Client(parser)
    cli.start()


if __name__ == '__main__':
    main()



#!/usr/bin/env python

import _socket
import logging
from argparse import ArgumentParser
from json import loads
from socket import AF_INET, SOCK_STREAM, SOL_SOCKET, SO_REUSEADDR, SO_REUSEPORT
from socket import error as SocketError, timeout as TimeoutError
from socket import socket
from threading import Lock, Thread
from typing import Dict
from uuid import uuid4

import daemon

from .config import SERVER_ADDR, TIMEOUT
from .network import Flag, Packet, send_pkt, recv_pkt
from .transfer import Sender, Receiver, Porter


class WatchDog(Thread):
    def __init__(self, server: 'Server', sock: socket):
        super().__init__(daemon=True)
        self.server = server
        self.sock = sock

    def run(self):
        try:
            # 等待接收新连接的第一个数据报文
            logging.debug('[WatchDog] waiting for handshake from %s:%d'
                          % self.sock.getpeername())
            self.sock.settimeout(60)
            packet = recv_pkt(self.sock)
            self.sock.settimeout(TIMEOUT)
        except ConnectionResetError:
            logging.error('[WatchDog] connection reset by peer.')
            return
        except TimeoutError:
            # 超时退出
            logging.error('[WatchDog] handshake timeout.')
            self.sock.close()
            return

        if packet.flag == Flag.PULL or packet.flag == Flag.PUSH:
            # 创建 Porter
            conn_info, = packet.unpack_body()
            porter = self.server.create_porter(packet.flag, conn_info)
            porter.conn_pool.add(self.sock)
            porter.start()

            # 将 SID 发送给客户端
            packet = Packet.load(Flag.SID, porter.sid)
            send_pkt(self.sock, packet)

        elif packet.flag == Flag.ATTACH:
            sid, = packet.unpack_body()
            if not self.server.porters[sid].conn_pool.add(self.sock):
                self.sock.close()

        else:
            # 对于错误的类型，直接关闭连接
            logging.debug('close conn')
            self.sock.close()


class Server(Thread):
    max_tasks = 256  # 同时运行的最大任务数量

    def __init__(self, max_conn) -> None:
        super().__init__(daemon=True)
        self.addr = SERVER_ADDR
        self.max_conn = max_conn  # 一个 Porter 的最大连接数
        self.is_running = True
        self.mutex = Lock()
        self.porters: Dict[bytes, Porter] = {}

    def create_porter(self, cli_flag: Flag, conn_info: str) -> Porter:
        '''创建新 Porter'''
        sid = uuid4().bytes
        _info = loads(conn_info)
        username = _info['user']
        if cli_flag == Flag.PULL:
            srcs = _info['srcs']
            include = _info['include']
            exclude = _info['exclude']
            logging.debug(f'[Server] New task-{sid.hex()} for send {srcs}')
            self.porters[sid] = Sender(sid, username, srcs, self.max_conn,
                                       include, exclude)
        else:
            dst_path = _info['dst']
            logging.debug(f'[Server] New task-{sid.hex()} for recv {dst_path}')
            self.porters[sid] = Receiver(sid, username, dst_path, self.max_conn)
        return self.porters[sid]

    def close_all_porters(self):
        '''关闭所有 Porter'''
        logging.debug('[Server] Closing all porters.')
        for porter in self.porters.values():
            porter.close()

    @staticmethod
    def create_socket_server(address, *, family=AF_INET, backlog=None,
                             reuse_port=False):
        """copyed from socket.py"""
        if reuse_port and not hasattr(_socket, "SO_REUSEPORT"):
            raise ValueError("SO_REUSEPORT not supported on this platform")

        sock = socket(family, SOCK_STREAM)
        try:
            if hasattr(_socket, 'SO_REUSEADDR'):
                try:
                    sock.setsockopt(SOL_SOCKET, SO_REUSEADDR, 1)
                except SocketError:
                    # Fail later on bind(), for platforms which may not
                    # support this option.
                    pass
            if reuse_port:
                sock.setsockopt(SOL_SOCKET, SO_REUSEPORT, 1)
            try:
                sock.bind(address)
            except SocketError as err:
                msg = '%s (while attempting to bind on address %r)' % \
                    (err.strerror, address)
                raise SocketError(err.errno, msg) from None
            if backlog is None:
                sock.listen()
            else:
                sock.listen(backlog)
            return sock
        except SocketError:
            sock.close()
            raise

    def run(self):
        self.srv_sock = self.create_socket_server(self.addr,
                                                  backlog=2048,
                                                  reuse_port=True)
        logging.info('[Server] Listening to %s:%d' % self.addr)
        while self.is_running:
            # wait for new connection
            cli_sock, cli_addr = self.srv_sock.accept()
            logging.info('[Server] Accept new connection: %s:%s' % cli_addr)

            # create a WatchDog for handshake
            dog = WatchDog(self, cli_sock)
            dog.start()


def main():
    parser = ArgumentParser()
    parser.add_argument('-d',
                        dest='daemon',
                        action='store_true',
                        help='daemonize the fcp process.')

    parser.add_argument('-c',
                        dest='concurrency',
                        metavar='NUM',
                        type=int,
                        default=128,
                        help='max concurrent connections of one task.')

    parser.add_argument('--loglevel',
                        metavar='LEVEL',
                        default='error',
                        choices=['debug', 'info', 'warning', 'error'],
                        help=('specify the server verbosity level. '
                              'Choices: debug | info | warning | error'))

    args = parser.parse_args()

    loglevel = getattr(logging, args.loglevel.upper())
    logformat = '%(asctime)s %(levelname)7s %(module)s.%(lineno)s: %(message)s'

    if args.daemon:
        with daemon.DaemonContext():
            logging.basicConfig(filename='/tmp/fcpd.log',
                                level=loglevel,
                                datefmt='%Y-%m-%d %H:%M:%S',
                                format=logformat)
            server = Server(args.concurrency)
            server.start()
            server.join()
    else:
        logging.basicConfig(level=loglevel,
                            datefmt='%Y-%m-%d %H:%M:%S',
                            format=logformat)
        server = Server(args.concurrency)
        server.start()
        server.join()


if __name__ == '__main__':
    main()
